# -*- coding: utf-8 -*-

# arteconcert.py
# This file is part of qarte-4
#    
# Author: Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
# Copyright: 2011-2020 Vincent Vande Vyvre
# Licence: GPL3
# Home page: https://launchpad.net/qarte
#
# arte Concert main class

# Need libqt5multimedia5-plugins

import os
import shutil
import json
import urllib.request
from urllib.error import URLError
from html.parser import HTMLParser
import html
import logging
lgg = logging.getLogger(__name__)

from datetime import datetime, timedelta
from threading import Thread
 
from PyQt5.QtGui import QFont, QTextCharFormat
from PyQt5.QtCore import (QObject, QCoreApplication, pyqtSignal, QLocale, QSize,
                          QTimer)
from PyQt5.QtMultimedia import QMediaPlayer

from loader import Loader
from tvdownloadingconfig import TvDownloadingConfig
from gui.warning import WarningBox
from data import USER_AGENT
from gui.videoplayer import VideoPlayer


class ArteConcert(QObject):
    """Define the arte Concert server handler.

    """
    parsingFinished = pyqtSignal()
    concertsCompleted = pyqtSignal()
    def __init__(self, core):
        super().__init__()
        self.core = core
        self.ui = core.ui
        self.page = self.ui.live_pg
        self.viewer = self.ui.live_pg.viewer
        self.lang = core.cfg['lang']
        self.default_quality = core.cfg['live_quality']
        self.fatal_error = False
        self.cache = core.workspace['live_tmp']
        self.concerts = []
        self.to_download = []
        self.categories = []
        self.video_items = {}
        self.is_loading = False
        self.loading_aborted = False
        self.current_video = False
        lang = QLocale.system().name()
        self.locale = QLocale(lang)
        # The next line crash with some graphics driver
        # bug: https://bugs.launchpad.net/bugs/1813131
        self.player = VideoPlayer(self, self.page.player_wdg)
        self.player.positionChanged.connect(self.on_player_position_changed)
        
        self.concertsCompleted.connect(self.show_recents)
        self.page.filter_led.commandTriggered.connect(self.search_and_find)
        self.page.start_btn.clicked.connect(self.preview_current_video)
        self.page.pause_btn.clicked.connect(self.pause_preview)
        self.page.stop_btn.clicked.connect(self.stop_preview)
        self.page.level_sld.valueChanged.connect(self.set_sound_level)
        self.page.player_wdg.wheelEvent = self.on_video_wheel_event

        self.loader = Loader()
        self.loader.loadingProgress.connect(self.display_progress)
        self.loader.loadingFinished.connect(self.on_downloading_finished)

    def config_parser(self):
        """Configure the main pages parser.

        """
        self.update_concerts()

        if not self.concerts:
            #TODO show warning dialog
            print("!!!!!! No concerts !!")
            return
        self.page.configure(self.concerts["categories"])
        for button in self.page.category_buttons:
            button.triggered.connect(self.change_category)

        self.parsingFinished.emit()

    def update_concerts(self):
        fpath = self.core.workspace['concerts_%s' % self.lang]
        cnt = self.get_update()
        self.concerts = []
        if cnt:
            try:
                with open(fpath, 'wb') as outf:
                    outf.write(cnt.encode('utf8', 'ignore'))
            except Exception as why:
                lgg.warning("Unable to write the arte concerts list")
                lgg.warning("Reason: %s" % why)

            try:
                self.concerts =  json.loads(cnt)
                lgg.info('Concert list updated')
            except Exception as why:
                lgg.warning("Unable to read the concerts list")
                lgg.warning("Reason: %s" % why)

        if not self.concerts:
            lgg.info('Use old data')
            self.get_old_concerts()

    def get_update(self):
        url = 'http://www.oqapy.eu/releases/concerts_%s.jsn' % self.lang
        req = urllib.request.Request(url, data=None,
                                     headers={"User-Agent": USER_AGENT})
        lgg.info('Get update ...')
        try:
            content = urllib.request.urlopen(req, timeout=5)
            return str(content.read().decode('utf-8', 'replace'))
        except Exception as why:
            lgg.info('urllib in update concert error: %s, %s' % (url, why))
            return False

    def get_old_concerts(self):
        self.concerts = {}
        fpath = self.core.workspace['concerts_%s' % self.lang]
        if os.path.isfile(fpath):
            try:
                with open(fpath, 'r') as objfile:
                    cnt = objfile.read()
                    self.concerts =  json.loads(cnt)
            except Exception as why:
                lgg.warning("Can't read concert_%s: %s" %(self.lang, why))

    def display_available_concerts(self):
        def build_items():
            cats = self.concerts["categories"]
            for key in cats:
                if key in ("categories", "indexes"):
                    continue
                if isinstance(self.concerts[key], list):
                    self.video_items[key] = []

                elif isinstance(self.concerts[key], dict):
                    keys = self.concerts[key].keys()
                    self.video_items[key] = {}
                    for k in keys:
                        d = self.concerts[key][k]
                        if isinstance(d, list):
                            self.video_items[key][k] = []
                            for x in d:
                                self.video_items[key][k].append(ConcertItem(x, key))

                        elif isinstance(d, dict):
                            self.video_items[key][k] = ConcertItem(d, key)

                else:
                    lgg.info("Unknow Item: %s" % key)
            self.concertsCompleted.emit()
        Thread(target=build_items).start()

    def show_recents(self):
        keys = sorted(self.video_items.keys())
        if "Playlists" in keys:
            self.page.set_active_category(keys.index("Playlists"), True)

        else:
            self.page.set_active_category(0, True)

    def change_category(self, category):
        keys = sorted(self.video_items.keys())
        self.show_category(keys[category])
        self.page.set_active_category(category)

    def show_category(self, name):
        lgg.info("show category: %s" % name)
        category = self.video_items[name]
        self.viewer.clear()
        keys = sorted(category.keys())
        for key in keys:
            if isinstance(category[key], list):
                th = "medias/playlist.png"
                title = key

            elif isinstance(category[key], ConcertItem):
                self.check_thumbnails(category[key])
                th = category[key].preview
                title = category[key].title

            link = "%s--%s" %(name, key)
            self.viewer.add_item(th, title, link=link)

        self.viewer.clean()

    def show_sublist(self, sublist, name):
        QCoreApplication.processEvents()
        self.viewer.clearSelection()
        self.viewer.clear()
        QCoreApplication.processEvents()
        for v in sublist:
            if isinstance(v, list):
                continue

            self.check_thumbnails(v)
            th = v.preview
            title = v.title
            link = "%s--%s" % (name, title)
            self.viewer.add_item(th, title, link=link)
            QCoreApplication.processEvents()

        self.page.enable_current_button()
        # Prevent a rending issue in QListWidget
        timer = QTimer.singleShot(200, self.viewer.select_first)
        QCoreApplication.processEvents()
            
    def concert_selection_changed(self, current):
        if self.viewer.currentItem() is None:
            return

        link = self.viewer.currentItem().link
        if link is None:
            self.search_by_id()
            return

        lgg.info("selected: %s" % link)
        path = link.split("--")
        if len(path) == 2:
            video = self.video_items[path[0]][path[1]]
            if isinstance(video, list):
                lgg.info("Show sublist of %s items" % len(video))
                self.show_sublist(video, link)

            elif isinstance(video, ConcertItem):
                lgg.info("Video %s, %s selected" % (video.title, video.order))
                self.display_concert_infos(video)

        elif len(path) == 3:
            for video in self.video_items[path[0]][path[1]]:
                if video.title == path[2]:
                    break
            lgg.info("Video %s, %s selected" % (video.title, video.order))
            self.display_concert_infos(video)

        self.current_video = video

    def check_thumbnails(self, item):
        if item.preview is not None:
            if os.path.isfile(item.preview):
                return

        try:
            ext = os.path.splitext(item.imgurl)[1] or ".jpg"
        except AttributeError:
            item.preview = "medias/noPreview.png"
            return

        dest = os.path.join(self.cache, "%s%s" %(item.order, ext))
        if not os.path.isfile(dest):
            item.preview = self.fetch_thumbnail(dest, item.imgurl)

        else:
            item.preview = dest

    def fetch_thumbnail(self, fname, url):
        for i in range(2):
            try:
                with open(fname, 'wb') as objfile:
                    f = urllib.request.urlopen(url)
                    objfile.write(f.read())
                return fname
            except Exception as why:
                lgg.info("Can't load image: %s" % url)
                lgg.info("Reason: %s" % why)
                # API error code 104 'Connection reset by peer'
                if not 'Errno 104' in why:
                    break

        return "medias/noPreview.png"

    def display_concert_infos(self, concert):
        lgg.info("Display info of %s" % concert.title)
        self.page.show_preview(concert.preview)
        self.page.file_name_led.setText(concert.title)
        self.show_pitch(concert)
        QCoreApplication.processEvents()
        self.search_available_videos(concert)

    def search_available_videos(self, item):
        """Return the available versions as a dict.

        Args:
        item -- ConcertItem instance
        """
        # Quick method first
        idx = item.order
        #https://www.w3schools.com/tags/ref_urlencode.asp
        url = "".join(["https://api.arte.tv/api/",
                      "player/v1/config/{0}/{1}?autostart=1&",
                      "lifeCycle=1&lang={2}_{3}"])
        url = url.format(self.lang, idx, self.lang, self.lang.upper())
        item.jsonurl = url
        content = self.get_page(url)

        if not content:
            # Failed, try second method
            content = self.get_page(item.url)
            p = Parser()
            p.feed(content)
            url = p.link
            if not url:
                lgg.info("json URL not found")
                return

            content = self.get_page(url)

        if not content:
            lgg.info("json URL unreadable")
            return

        try:
            dct = json.loads(content)
            vers = dct["videoJsonPlayer"]
        except Exception as why:
            lgg.info("json read error: %s" % why)
            return

        # Debugging only
        #with open(item.title, "wb") as objf:
            #jsn = json.dumps(vers, sort_keys=True, indent=4, 
                         #separators=(',', ': '), ensure_ascii=False)
            #objf.write(jsn.encode('utf-8', 'replace'))

        desc = ""
        if "V7T" in vers:
            desc += vers["V7T"] + "\n\n"

        if "VDE" in vers:
            desc += vers["VDE"]

        if desc:
            title = item.title
            date = "%s %s" %(item.format_date(), item.format_duration())
            self.update_pitch(title, date, desc)

        if "VSR" in vers:
            c = vers["VSR"]
            if c:
                item.qualities = self.clean_qualities(c)
                self.show_available_qualities(item)
                return

        self.show_available_qualities("Sorry, video not available")

    def get_page(self, url):
        lgg.info('Load page: %s' % url)
        user_agent = 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:49.0)'\
                  ' Gecko/20100101 Firefox/49.0'
        req = urllib.request.Request(url, data=None, 
                                 headers={"User-Agent": user_agent})
        try:
            content = urllib.request.urlopen(req)
        except Exception as why:
            lgg.warning('urllib error: %s, %s' % (url, why))
            return False

        return str(content.read().decode('utf-8', 'replace'))

    def show_pitch(self, item):
        """Called when the current item in the viewer has changed.

        """
        title = item.title
        date = "%s %s" %(item.format_date(), item.format_duration())
        pitch = "No description"
        if item.short_desc != "Unknow" and item.short_desc is not None:
            pitch = item.short_desc
            if item.long_desc != "Unknow" and item.long_desc is not None:
                pitch = pitch + "<br /><br />" + item.long_desc

        elif item.desc != "Unknow" and item.desc is not None:
            pitch = item.desc

        self.update_pitch(title, date, pitch, html=True)
 
    def update_pitch(self, *args, html=False):
        """Show the pitch of the current item.

        """
        self.page.editor_live.clear()
        font = QFont("sans", 10)
        c_form = QTextCharFormat()
        c_form.setFont(font)
        font1 = QFont("sans", 12)
        font1.setWeight(75)
        c_form1 = QTextCharFormat()
        c_form1.setFont(font1)
        self.page.editor_live.setCurrentCharFormat(c_form1)
        self.page.editor_live.appendPlainText(args[0])
        self.page.editor_live.setCurrentCharFormat(c_form) 
        self.page.editor_live.appendPlainText(args[1])
        self.page.editor_live.appendPlainText("")
        if html:
            self.page.editor_live.appendHtml(args[2])

        else:
            self.page.editor_live.appendPlainText(args[2])

        self.page.editor_live.verticalScrollBar().setValue(0)

    def show_available_qualities(self, video):
        self.page.vers_twg.clear()
        self.page.vers_twg.setRowCount(0)
        if isinstance(video, str):
            self.page.vers_twg.show_warning(video)

        else:
            for key in sorted(video.qualities.keys()):
                self.page.vers_twg.add_item(key, video.qualities[key])

    def select_video_to_download(self, quality, basket=True):
        """Select a video for the downloading basket.

        Args:
        idx -- index of video if called by a double click on a item
        """
        video = self.current_video
        if video in self.to_download:
            return

        lgg.info("Append item {0} to download basket".format(video.order))

        video.set_file_name(self.page.file_name_led.text())
        video.set_quality(quality)
        if basket:
            self.to_download.append(video)
            self.page.live_basket.add_concert(video, self.page.pixmap)
            self.page.dl_live_wdg.dwnld_btn.setEnabled(True)
            self.ui.actionDo_wnload.setEnabled(True)

    def remove_item_from_basket(self, idx=None):
        """Called if user remove an item from downloading basket or when
        a downloading is complete.

        idx -- index of item
        """
        logger.info('Remove item %s frombasket' % idx)
        if idx == 'first':
            idx = 0

        elif idx is None:
            idx = self.get_current_item_index()

        self.page.live_basket.delete_item(idx)
        self.to_download.pop(idx)

    def get_stream_urls(self, item):
        content = self.get_page(item.jsonurl)
        if not content:
            lgg.info("Can't get the stream urls page: %s" % self.urlliberror)
            warn = WarningBox(5, why, self.ui)
            rep = warn.exec_()
            return

        try:
            dct = json.loads(content)
            strs = dct["videoJsonPlayer"]['VSR']
        except Exception as why:
            lgg.info("Can't read the stream urls: %s" % why)
            warn = WarningBox(5, why, self.ui)
            rep = warn.exec_()
            return

        if not strs:
            why = "Can't read the stream urls"
            try:
                msg = dct["videoJsonPlayer"]["custom_msg"]
                if msg["msg"].startswith(("Désolé", "Leider ist")):
                    why = msg["msg"]
            except:
                pass
            warn = WarningBox(5, why, self.ui)
            rep = warn.exec_()
            return

        return strs

    def clean_qualities(self, dct):
        keys = list(dct.keys())
        for key in keys:
            if not 'HTTP' in key:
                dct.pop(key)

        return dct

    def search_and_find(self):
        key = self.page.filter_led.text().lower()
        if len(key) < 4:
            lgg.info("Keyword too small: %s letters" % len(key))
            return

        def search(videos):
            if isinstance(videos, ConcertItem):
                if key in videos.title.lower():
                    candidates.append(videos)
            else:
                for video in videos:
                    if isinstance(video, str):
                        search(self.video_items[category][video])

                    elif isinstance(video, list):
                        search(video)

                    elif isinstance(video, dict):
                        for _, value in videos.items():
                            search(value)

                    elif isinstance(video, ConcertItem):
                        if key in video.title.lower():
                            candidates.append(video)

        candidates = []

        for category in self.video_items:
            search(self.video_items[category])

        if candidates:
            self.page.result_lbl.setText("Found %s concerts" % len(candidates))
            self.viewer.clear()
            for candidate in candidates:
                self.check_thumbnails(candidate)
                th = candidate.preview
                self.viewer.add_item(th, candidate.title, idx=candidate.order)

        else:
            self.page.result_lbl.setText("Not found")

        timer = QTimer.singleShot(5000, self.erase_result)

    def erase_result(self):
        self.page.result_lbl.setText("")

    def search_by_id(self):
        index = self.viewer.currentItem().index
        if index is None:
            return

        def search(videos):
            if isinstance(videos, ConcertItem):
                if index == videos.order:
                    self.item = videos
                    return

            else:
                for video in videos:
                    if isinstance(video, str):
                        search(self.video_items[category][video])
                        if self.item:
                            break

                    elif isinstance(video, list):
                        search(video)

                    elif isinstance(video, ConcertItem):
                        if video.order == index:
                            self.item = video
                            break

        self.item = False
        for category in self.video_items:
            search(self.video_items[category])
            if self.item:
                break

        if self.item:
            self.display_concert_infos(self.item)
            self.current_video = self.item

        else:
            lgg.warning("Unable to find the item %s" % index)

    def on_empty_list(self):
        self.page.dl_live_wdg.dwnld_btn.setEnabled(False)
        self.ui.actionDo_wnload.setEnabled(False)
        self.ui.actionCa_ncel.setEnabled(False)

    def edit_file_name(self, text=None):
        if text is None:
            text = self.page.file_name_led.text()

        lgg.info("Change file name: %s" % text)
        if text:
            self.current_video.set_file_name(text)

            # Refresh basket if needed
            if self.current_video in self.to_download:
                idx = len(self.to_download) - 1
                self.page.live_basket.rename(idx, text)

    def configure_downloading(self, video):
        """Call the downloading config dialog box.

        Args:
        video -- ConcertItem instance
        """
        lgg.info('Configure downloading of %s' % video.title)
        if video.qualities is None:
            self.select_video_to_download(None, False)

        self.cfg = TvDownloadingConfig(self)
        self.cfg.configure(video.qualities, video, 
                           self.core.cfg['live_quality'], 
                           self.core.cfg['lang'], sort=False)
        self.cfg.exec_()

    def set_preferred_quality(self, quality):
        self.core.cfg['live_quality'] = quality

    def start_download(self):
        if not self.to_download:
            self.upage.dl_live_wdg.clear(True)
            return

        video = self.to_download.pop(0)
        lgg.info('Start download: %s' % video.title)
        stream = self.get_stream(video)
        target = self.core.get_target_path(self.core.cfg['music_folder'])
        self.video_loaded = self.get_video_path(video)
        if not self.initiate_loader(stream, target):
            return

        if not self.core.check_free_space(self.file_size, 
                                          self.core.cfg["music_folder"]):
            self.reset_loading_widgets()
            self.to_download.insert(0, video)
            return

        self.update_widgets_on_loading()
        lgg.info('Start download %s quality: %s file size: %s'
                    %(video.fname, video.quality, self.file_size))
        lgg.info('URL: %s' % stream)
        self.is_loading = True
        Thread(target=self.loader.download_mp4).start()

    def get_stream(self, item):
        return item.qualities[item.quality]["url"]

    def initiate_loader(self, url, fname):
        self.file_size = self.loader.config_loader(url, fname)
        if isinstance(self.file_size, int):
            return True

        lgg.info('Loading error: %s' % self.file_size)
        return False

    def update_widgets_on_loading(self):
        self.page.live_basket.disable_settings()
        self.page.dl_live_wdg.on_progress()
        self.ui.actionDo_wnload.setEnabled(False)
        self.ui.actionCa_ncel.setEnabled(True)

    def reset_loading_widgets(self):
        self.page.dl_live_wdg.dwnld_btn.setEnabled(True)
        self.page.dl_live_wdg.cancel_btn.setEnabled(False)
        self.ui.actionDo_wnload.setEnabled(True)
        self.ui.actionCa_ncel.setEnabled(False)

    def get_video_path(self, item):
        """Return the path/name of the video file.

        Args:
        item -- TVItem instance
        """
        fld = self.core.cfg['music_folder']
        path = os.path.join(fld, item.fname + ".mp4")
        count = 1
        while os.path.isfile(path):
            path = os.path.join(fld, item.fname + "(%s).mp4" % count)
            count += 1

        return path

    def cancel_download(self):
        lgg.info('Cancel download')
        self.loading_aborted = True
        self.loader.kill_request = True

    def display_progress(self, percent, speed, remain):
        """Show the downloading progress informations.

        Args:
        percent -- percent downloaded
        speed -- bandwidth in Ko/sec.
        remain -- estimated remaining time
        """
        rtime = int(remain / speed)
        speed = self.locale.toString(int(speed / 1024))
        self.page.dl_live_wdg.progressBar.setValue(percent)
        self.page.dl_live_wdg.progress_lbl.setText(
                                "Speed: %s Kio/sec." % speed)
        m, s = divmod(rtime, 60)
        self.page.dl_live_wdg.progress_lbl2.setText(
                                "Remaining time: %s min. %s sec." %(m, s))

    def on_downloading_finished(self, fname):
        """Rename the video downloaded.

        Args:
        fname -- temporary file name
        """
        lgg.info('File %s downloaded' % fname)
        self.is_loading = False
        item = self.page.live_basket.remove_item(0)
        self.page.dl_live_wdg.clear()
        self.ui.actionCa_ncel.setEnabled(False)
        if not fname:
            self.handle_downloading_error()
            return

        try:
            os.rename(fname, self.video_loaded)
            lgg.info('File renamed: %s' % os.path.basename(self.video_loaded))
        except Exception as why:
            lgg.warning("Can't rename %s" % self.video_loaded)
            lgg.warning("Reason: %s" % why)

        self.check_file_size()
        self.copy_summary(item, self.video_loaded)
        self.copy_thumbnail(item, self.video_loaded)
        if self.to_download:
            self.start_download()

    def check_file_size(self):
        if self.loading_aborted:
            # The downloading was stopped by the user
            self.loading_aborted = False
            return

        expected = self.loader.size
        loaded = os.path.getsize(self.video_loaded)
        if loaded != expected:
            # TODO format numbers
            lgg.warning("The downloading seems incomplete!")
            lgg.warning("Size expected: %s downloaded: %s" %(expected, loaded))
            fname = os.path.basename(self.video_loaded)
            warn = WarningBox(9, (fname, expected, loaded), self.ui)
            rep = warn.exec_()

    def copy_summary(self, video, fname):
        """Save the summary into a file.

        Args:
        video -- TVItem instance
        fname -- path-name of the video downloaded
        """
        if not self.core.cfg['pitch_live']:
            return

        if self.core.cfg['pitch_live_unique']:
            path = os.path.join(os.path.dirname(fname), 'index')

        else:
            path = os.path.splitext(fname)[0]

        title = video.fname
        _, date, summary = self.get_description(video)
        txt = "\n* %s\n  %s\n\n  %s\n" %(title, date, summary)
        try:
            with open(path, 'a') as outf:
                outf.write(txt)
                lgg.info("Summary saved")
        except Exception as why:
            lgg.warning("Can't write summary: %s" % path)
            lgg.warning("Reason: %s" % why)

    def copy_thumbnail(self, video, fname):
        """Save the thumbnail into a file.

        Args:
        video -- TVItem instance
        fname -- path-name of the video downloaded
        """
        if self.core.cfg['copythumb_live']:
            path = os.path.splitext(fname)[0] + ".jpg"
            try:
                orig = video.preview
                shutil.copy(orig, path)
                lgg.info("Thumbnail saved")
            except:
                return

    def handle_downloading_error(self):
        """Display a warning box when the downloading failed.

        """
        url = self.loader.url
        error = self.loader.last_error
        lgg.warning("Downloading error: %s" % error)
        lgg.warning("URL: %s" % url)
        warn = WarningBox(8, (url, error), self.ui)
        rep = warn.exec_()
        if self.page.live_basket.lst_movies:
            self.start_download()

    def get_description(self, item):
        title = item.title
        date = "%s %s" %(item.format_date(), item.format_duration())
        pitch = "No description"
        if item.short_desc != "Unknow" and item.short_desc is not None:
            pitch = item.short_desc
            if item.long_desc != "Unknow" and item.long_desc is not None:
                pitch = pitch + "<br /><br />" + item.long_desc

        elif item.desc != "Unknow" and item.desc is not None:
            pitch = item.desc

        return title, date, pitch

    def download_later(self, index):
        lgg.info('Call deferred task with item %s' % index)
        video = self.viewer.videos_items[index]
        qualities = self.get_stream_urls(video)

        if qualities is None:
            lgg.warning("No stream URLs found, item rejected")
            return

        self.clean_qualities(qualities)
        video.qualities = qualities
        video.set_file_name()
        video.set_quality(self.default_quality)
        self.core.create_deferred_task(video)

    #---------------------------------------------------------
    # Previewing video
    #---------------------------------------------------------
    def preview_current_video(self):
        if self.player.is_on_pause:
            self.page.set_running_player()
            self.player.start()
            return

        if not self.current_video:
            return

        if "HTTPS_EQ_1" in self.current_video.qualities:
            quality = self.current_video.qualities["HTTPS_EQ_1"]

        elif "HTTPS_HQ_1" in self.current_video.qualities:
            quality = self.current_video.qualities["HTTPS_HQ_1"]

        elif "HTTPS_SQ_1" in self.current_video.qualities:
            quality = self.current_video.qualities["HTTPS_SQ_1"]

        else:
            # Damn ! So, take one randomly.
            k = self.current_video.qualities.keys()[0]
            quality = self.current_video.qualities[k]

        url = quality["url"]
        self.time_ = datetime(2018, 1, 1, 0, 0, 0)
        self.player.set_media(url)
        self.page.thumbView.hide()
        self.page.player_wdg.show()
        self.page.set_running_player()
        d = self.current_video.duration
        if d is not None:
            delta = timedelta(seconds=d)
            n = self.time_ + delta
            t = "%02d:%02d:%02d" %(n.hour, n.minute, n.second)
            self.page.duration_lbl.setText(t)
        QCoreApplication.processEvents()
        self.player.start()

    def pause_preview(self):
        self.page.set_player_pause()
        self.player.set_on_pause()

    def stop_preview(self):
        self.player.finish()
        self.page.player_wdg.hide()
        self.page.thumbView.show()
        self.page.stop_player()

    def on_player_position_changed(self, position):
        d = timedelta(milliseconds=position)
        n = self.time_ + d
        t = "%02d:%02d:%02d" %(n.hour, n.minute, n.second)
        self.page.elapsed_lbl.setText(t)

    def set_sound_level(self, level):
        self.player.set_sound_level(level)
        self.page.level_lbl.setText(str(level))

    def on_video_wheel_event(self, event):
        lgg.info("Wheel event on player: %s" % event.angleDelta().y())
        level = event.angleDelta().y() / 24
        self.player.setVolume(level+self.page.level_sld.value())
        event.accept()

    def on_space_key_pressed(self):
        if self.player.is_running:
            if not self.player.is_on_pause:
                self.pause_preview()
                return True

        else:
            if self.player.is_on_pause:
                self.page.set_running_player()
                self.player.start()
                return True

        return False


class ConcertItem(object):
    def __init__(self, item, cat):
        """Item concert in arte Live Web.

        Args:
        cat -- category
        item -- values provided by the parser
        lang -- user langage ('fr' or 'de')
        """
        self.category = cat
        self.order = item["programId"]
        self.title = item["title"]
        self.date = ""
        self.expires = ""
        self.desc = item["description"]
        self.short_desc = item["shortDescription"]
        self.long_desc = item["fullDescription"]
        self.duration = item["duration"]
        self.url = item["url"]
        self.imgurl = item["pict"]
        self.jsonurl = ""
        self.preview = None
        self.qualities = None
        self.quality = None
        self.lang = ""
        self.fname = None
        self.credits = item["credits"]

    def set_file_name(self, new=False):
        if not new:
            new = self.title

        f = new.replace('/', '-').replace('»', '_').replace('«', '_')
        f = f.replace('\\', ' ').replace('.', ' ')
        if len(f) > 48:
            f = "%s---%s" %(f[:35], f[-10:])

        self.fname = f

    def set_quality(self, quality):
        if not quality in self.qualities:
            for q in ['HTTP_EQ_1', 'HTTP_HQ_1', 'HTTP_SQ_1', 'HTTP_MQ_1',
                      'HTTPS_EQ_1', 'HTTPS_HQ_1', 'HTTPS_SQ_1', 'HTTPS_MQ_1']:
                if q in self.qualities:
                    quality = q
                    break

        self.quality = quality

    def get_image(self):
        if self.preview is not None:
            try:
                img = urllib2.urlopen(self.pix).read()
            except Exception as why:
                return None

            else:
                return img

    def format_date(self):
        return self.date.replace('T', ' ').rstrip('Z')

    def format_duration(self):
        if self.duration is None:
            return ""

        min_, sec = divmod(int(self.duration), 60)
        if min_ < 60:
            return "%s min. %s sec." %(min_, sec)

        hours, min_ = divmod(min_, 60)
        return "%s h. %s min. %s sec." %(hours, min_, sec)


class Parser(HTMLParser):
    """Define the parser of arte Concert's pages.
<div class="video-embed"><iframe allowfullscreen="true" style="transition-duration:0;transition-property:no;margin:0 auto;position:relative;display:block;background-color:#000000;" frameborder="0" scrolling="no" width="100%" height="100%" src="https://www.arte.tv/player/v3/index.php?json_url=https%3A%2F%2Fapi.arte.tv%2Fapi%2Fplayer%2Fv1%2Fconfig%2Ffr%2F076805-003-A%3Fautostart%3D1%26lifeCycle%3D1&lang=fr_FR"></iframe>
    """
    def __init__(self):
        super().__init__()
        self.link = False
        self.inframe = False

    def handle_starttag(self, tag, attrs):
        if tag == "iframe":
            self.inframe = True
            for att in attrs:
                if att[0] == 'src':
                    self.link = att[1]

    def handle_endtag(self, tag):
        if tag == "iframe":
            self.inframe = False



# Devel usage ------------------------------------------------------------------
class Core:
    def __init__(self):
        self.ui = None
        self.cfg = {'lang': 'fr'}
        ws = os.path.expanduser("~") + "/qarte-3/wspace/livePreviews"
        self.workspace = {'live_tmp': ws}

        self.arteconcert = ArteConcert(self)
        self.arteconcert.config_parser()

if __name__ == '__main__':
    core = Core()
