# -*- coding: utf-8 -*-

# basket.py
# This file is part of qarte-4
#    
# Author: Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
# Copyright: 2011-2020 Vincent Vande Vyvre
# Licence: GPL3
# Home page: https://launchpad.net/qarte
#
# Dowloading basket

from PyQt5.QtCore import (Qt, pyqtSignal, QSize, QMimeData, QPoint, QTimer,
                          QCoreApplication, QItemSelectionModel, )
from PyQt5.QtGui import (QIcon, QPixmap, QDrag, )
from PyQt5.QtWidgets import (QTableWidget, QAbstractItemView, QApplication,
                             QTableWidgetItem, QToolButton, )


class Basket(QTableWidget):
    itemAdded = pyqtSignal(int)
    def __init__(self, ui, parent=None):
        super().__init__(parent)
        self.ui = ui
        self.core = ui.core
        self.driver = None
        self.lst_movies = []
        self.setMaximumSize(QSize(350, 16777215))
        self.setShowGrid(False)
        self.setDragEnabled(True)
        self.setDragDropMode(QAbstractItemView.InternalMove)
        self.setDefaultDropAction(Qt.CopyAction)
        self.setDropIndicatorShown(True)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.verticalHeader().hide()
        self.horizontalHeader().hide()
        self.setIconSize(QSize(85, 48))
        self.setColumnCount(4)
        self.setColumnWidth(0, 85)
        self.setColumnWidth(1, 165)
        self.setColumnWidth(2, 24)
        self.setColumnWidth(3, 24)

    def mousePressEvent(self, event):
        self.idx = False
        mod = QApplication.keyboardModifiers()
        idx = self.indexAt(event.pos())
        btn = event.button()
        if btn == 1 and not idx.isValid():
            self.clearSelection()

        elif btn == 1 and idx.isValid():
            self.idx = idx

        QTableWidget.mousePressEvent(self, event)

    def mouseMoveEvent(self, event):
        if self.idx:
            # Don't move the item 0 if is loading
            if not self.idx.row() and self.driver.is_loading:
                return

            self.startDrag(event)

    def startDrag(self, event):
        index = self.indexAt(event.pos())
        mimeData = QMimeData()
        mimeData.setText("video")
        drag = QDrag(self)
        drag.setMimeData(mimeData)
        pixmap = QPixmap()
        pixmap = self.item(index.row(), 0).icon().pixmap(80, 60, 
                                                        QIcon.Normal, 
                                                        QIcon.Off)
        drag.setPixmap(pixmap)
        drag.setHotSpot(QPoint(pixmap.width()/2, pixmap.height()))
        drag.setPixmap(pixmap)
        result = drag.exec_(Qt.MoveAction)
        event.accept()

    def dragEnterEvent(self, event):
        if event.mimeData().hasText():
            event.accept()

        else:
            event.ignore()

    def dragMoveEvent(self, event):
        idx = self.indexAt(event.pos())
        if idx.isValid() and event.mimeData().hasText():
            event.setDropAction(Qt.MoveAction)
            self.highlight_item(self.indexAt(event.pos()))

        elif str(event.mimeData().text()).startswith('new'):
            event.setDropAction(Qt.MoveAction)

        else:
            event.ignore()

    def dragLeaveEvent(self, event):
        pass

    def dropEvent(self, event):
        idx = self.indexAt(event.pos())
        if not event.mimeData().hasText():
            event.ignore()
            return

        data = str(event.mimeData().text())
        if data == "video" and idx.isValid() and idx != self.idx:
            self.move_item_to(self.idx.row(), idx.row())

        elif data.startswith('new'):
            self.add_object(int(data.split(' ')[1]))

    def move_item_to(self, src, dest):
        if not dest and self.driver.is_loading:
            # Don't shift the item 0 if is loading
            if src == 1:
                return

            dest == 1

        ic = self.item(src, 0).icon()
        txt = self.item(src, 1).text()
        video = self.remove_item(src)
        self.insertRow(dest)
        self.setRowHeight(dest, 48)
        item = QTableWidgetItem()
        item.setIcon(ic)
        self.setItem(dest, 0, item)
        item = QTableWidgetItem()
        item.setText(txt)
        self.setItem(dest, 1, item)
        item = TableButton(dest, 'settings', self)
        self.setCellWidget(dest, 2, item)
        item = TableButton(dest, 'remove', self)
        self.setCellWidget(dest, 3, item)

        self.update_buttons()
        self.lst_movies.insert(dest, video)

    def remove_item(self, row=None):
        """Remove one item from table and from list.

        """
        if row is None:
            row = self.rowCount() - 1
        self.removeRow(row)
        video = self.lst_movies.pop(row)

        self.clearSelection()
        self.update_buttons()
        if not self.rowCount():
            self.driver.on_empty_list()

        return video

    def is_empty(self):
        return not self.rowCount()

    def print_titles(self):
        """Debugging function"""
        for v in self.lst_movies:
            print(v.title)

    def highlight_item(self, idx):
        self.clearSelection()
        self.selectionModel().setCurrentIndex(idx, 
                                QItemSelectionModel.Select)

    def update_buttons(self):
        for r in range(self.rowCount()):
            # There's no item in column 2 for the concerts basket
            if self.cellWidget(r, 2) is not None:
                self.cellWidget(r, 2).row_ = r
            self.cellWidget(r, 3).row_ = r

    def config_download(self, idx):
        self.driver.configure_downloading(self.lst_movies[idx])

    def add_concert(self, video, pixmap):
        """Add an item concert into the basket.

        Args:
        video -- instance of ConcertItem
        pixmap -- the image used into the viewer
        """
        text = video.fname
        if text is None:
            text = video.title

        text = self.format_text(text)
        r = self.rowCount()
        self.insertRow(r)
        self.setRowHeight(r, 48) 
        item = QTableWidgetItem()
        item.setIcon(QIcon(pixmap))
        self.setItem(r, 0, item)
        item = QTableWidgetItem()
        item.setText(text)
        self.setItem(r, 1, item)
        item = TableButton(r, 'remove', self)
        self.setCellWidget(r, 3, item)
        self.lst_movies.append(video)

    def add_object(self, idx, settings=True):
        """Add a video into the basket.

        Args:
        idx -- index of video
        settings -- if True, display the button settings.
                    Not used in arte concerts
        """
        if self.driver.is_in_category:
            video = self.ui.arte_list.video_items[idx].item

        else:
            video = self.driver.videos[idx]

        if video in self.lst_movies:
            self.ui.arte_list.clearSelection()
            return

        def send_signal():
            self.itemAdded.emit(idx)

        img = QPixmap(video.pixmap)
        text = self.format_text(video.fname)
        r = self.rowCount()
        self.insertRow(r)
        self.setRowHeight(r, 48) 
        item = QTableWidgetItem()
        item.setIcon(QIcon(img))
        self.setItem(r, 0, item)
        item = QTableWidgetItem()
        item.setText(text)
        self.setItem(r, 1, item)
        if settings:
            item = TableButton(r, 'settings', self)
            self.setCellWidget(r, 2, item)
        item = TableButton(r, 'remove', self)
        self.setCellWidget(r, 3, item)

        self.lst_movies.append(video)
        QCoreApplication.processEvents()
        self.ui.tv_basket_buttons.dwnld_btn.setEnabled(True)
        self.ui.action_Download.setEnabled(True)
        #self.ui.action_Download.setEnabled(True)
        self.ui.arte_list.clearSelection()
        timer = QTimer()
        timer.singleShot(300, send_signal)
        timer.start()

    def rename(self, row, txt):
        """Rename an item.

        Args:
        row -- row
        txt -- new name
        """
        self.item(row, 1).setText(txt)

    def format_text(self, txt):
        """Format the movie's title.

        Args:
        txt -- title

        Returns:
        Formated text
        """
        if len(txt) < 31:
            return txt

        form = ""
        while 1:
            chain = txt[:31]
            count = len(chain) - 1
            while 1:
                if chain[count]  in [" ", "_", "-"]:
                    break

                count -= 1
                if count == 0:
                    count = len(chain) - 1
                    break

            form = form + chain[:count] + "\n"
            txt = txt[count+1:]
            if not txt:
                break

            elif len(txt) < 31:
                form = form + txt + "\n"
                break

        return form[:-1]

    def disable_settings(self):
        """Disable the settings for the video that is downloading.

        """
        try:
            self.cellWidget(0, 3).setEnabled(False)
            self.cellWidget(0, 2).setEnabled(False)
        except:
            pass

class TableButton(QToolButton):
    def __init__(self, row, feature, parent=None):
        super().__init__()
        self.row_ = row
        self.feature = feature
        self.table = parent
        self.setIconSize(QSize(20, 20))
        self.set_icon()
        self.clicked.connect(self.send_signal)
        self.setStyleSheet("""
                QToolButton {background-color: qlineargradient(
                    x1: 0, y1: 0, x2: 0, y2: 1,
                    stop: 0 #FFFFFF, stop: 0.4 #FAFAFA,
                    stop: 0.5 #FDFDFD, stop: 1.0 #F0F0F0);
                    border-radius: 4px;
                    padding: 2px;}
                QToolButton:hover {border: 2px solid #CC1D00;}
                QToolButton:pressed {background-color: qlineargradient(
                    x1: 0, y1: 0, x2: 0, y2: 1,
                    stop: 0 #000000, stop: 0.4 #050505,
                    stop: 0.5 #606060, stop: 1.0 #101010);}""")

    def set_icon(self):
        icon = QIcon()
        if self.feature == 'settings':
            pix = QPixmap('medias/settings.png')
        else:
            pix = QPixmap('medias/list-remove.png')
        icon.addPixmap(pix, QIcon.Normal, QIcon.Off)
        self.setIcon(icon)

    def send_signal(self, b):
        if self.feature == 'remove':
            self.table.remove_item(self.row_)
        else:
            self.table.config_download(self.row_)
