# -*- coding: utf-8 -*-

# loadingscheduler.py
# This file is part of qarte-4
#    
# Author: Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
# Copyright: 2016-2018 Vincent Vande Vyvre
# Licence: GPL3
# Home page: https://launchpad.net/qarte
#
# Deferred downloading scheduler

import sys
import os
import time
import pwd
import json
import pickle
import locale
import logging
lgg = logging.getLogger(__name__)

from datetime import datetime
from threading import Thread

from PyQt5.QtWidgets import QApplication

from gui.schedulerdialog import SchedulerDialog
from pycrontab import PyCronTab

class LoadingScheduler:
    def __init__(self, core, video=None):
        self.core = core
        self.ui = core.ui
        self.video = video
        self.lang = core.cfg['lang']
        self.logfile = os.path.join(core.workspace['user'], 'daemon.log')
        self.cron_cmd = core.cfg['crontab_cmd']
        self.expired = []
        self.edit_tasks()
        self.load_task_file()
        self.dialog = SchedulerDialog(self, core.ui)
        if video:
            self.set_video_item(video)
        self.dialog.log_editor.set_log_filename(self.logfile)
        self.dialog.log_editor.load_log()
        self.dialog.show()
        self.show_tasks()
        rep = self.dialog.exec_()
        
    def create_task(self):
        if self.dialog.arteconcert_rdo.isChecked():
            self.create_concert_task()
            return

        order = self.get_new_order()
        dial = self.dialog
        new = {}
        if dial.fname_rdo.isChecked():
            new['title'] = self.video.title
            new['fname'] = self.sanitize_file_name(dial.fname_led.text())
            new['stream'] = self.video.streams[self.video.quality]["url"]
            new['quality'] = self.video.quality
            new['type'] = 0

        elif dial.serial_rdo.isChecked():
            name = dial.serials_cmb.currentText()
            if 'GEO' in name:
                new['kwords'] = '360 geo'
                new['type'] = 2

            else:
                new['title'] = dial.serials_cmb.currentText()
                new['type'] = 1
        else:
            kwords = self.get_keywords(dial.kwords_led.text())
            if not kwords:
                return

            new['kwords'] = kwords
            new['type'] = 2

        new['all'] = False
        if new['type']:
            new['all'] = dial.all_rdo.isChecked()

        if dial.date_rdo.isChecked():
            date, time = self.get_task_date()
            if not self.check_datetime(date, time):
                dial.set_invalid_date_warning()
                return

            new['date'] = (date.day(), date.month(), date.year())
            new['time'] = (time.hour(), time.minute())

        else:
            new['date'] = dial.week_cmb.currentIndex()
            time = dial.time_spb_2.time()
            new['time'] = (time.hour(), time.minute())

        self.tasks[order] = new
        self.save_task_file()
        self.register_task(order, new)

#{order: {title: bla,
#         fname: file/name, no file name with serial and kwords
#         type: 0=video name, 1=serial name, 2=kwords, 3=concert
#         all: True, False
#         date: date as tuple(D, M, Y) or weekly
#         time: time as tuple(H, M), None
#         mp4url: http:........,
#         quality: if not mp4url,
#         kwords: []}}

    def create_concert_task(self):
        order = self.get_new_order()
        dial = self.dialog
        new = {}
        new['title'] = self.video.title
        new['fname'] = self.sanitize_file_name(dial.fname_led.text())
        new['stream'] = self.video.qualities[self.video.quality]["url"]
        new['type'] = 3
        self.tasks[order] = new
        date, time = self.get_task_date()
        if not self.check_datetime(date, time):
            dial.set_invalid_date_warning()
            return

        new['date'] = (date.day(), date.month(), date.year())
        new['time'] = (time.hour(), time.minute())
        self.tasks[order] = new
        self.save_task_file()
        self.register_task(order, new)

    def get_task_date(self):
        date = self.dialog.date_spb.date()
        time = self.dialog.time_spb.time()
        return date, time

    def get_stream_url(self):
        quality = self.dialog.quality_cmb.currentText()
        return quality

    def load_task_file(self):
        fld = self.core.workspace['user']
        path = os.path.join(fld, 'tasks.json')
        self.tasks = False
        if os.path.isfile(path):
            lgg.info('Load task file: %s' % path)
            try:
                with open(path, 'r') as objfile:
                    cnt = objfile.read()
                    self.tasks = json.loads(cnt)
            except Exception as why:
                lgg.warning("Read error from: {0},".format(path))
                lgg.warning("reason: {0}.".format(why))

        if self.tasks:
            self.clean_tasks()

        else:
            self.tasks = {}

    def save_task_file(self):
        fld = self.core.workspace['user']
        path = os.path.join(fld, 'tasks.json')
        lgg.info('Save task file: %s' % path)
        jsn = json.dumps(self.tasks, sort_keys=True, indent=4, 
                         separators=(',', ': '), ensure_ascii=False)
        try:
            with open(path, 'wb') as outfile:
                outfile.write(jsn.encode('utf-8', 'replace'))
        except Exception as why:
            lgg.warning("Write error: {0}".format(path))
            lgg.warning("reason: {0}".format(why))

    def set_video_item(self, video):
        self.video = video
        self.dialog.configure(video)

    def get_new_order(self):
        if not self.tasks:
            return '001'

        last = sorted(self.tasks.keys())[-1]
        return "%03d" % (int(last) + 1)

    def sanitize_file_name(self, name):
        return name

    def get_keywords(self, line):
        return [i.lower() for i in line.split(' ')]

    def check_datetime(self, d, t):
        """Check if date and time of cron are not too short.

        Args:
        d -- QDate
        t -- QTime

        Returns:
        False if datetime is too short, True otherwise
        """
        now = datetime.now()
        dt = datetime(d.year(), d.month(), d.day(), t.hour(), t.minute(), 0)
        if now >= dt:
            return False

        return True

    def clean_tasks(self):
        expired = []
        for key, value in self.tasks.items():
            if isinstance(value['date'], int):
                # Permanent task
                continue
            d, m, y = value['date']
            h, mn = value['time']
            if datetime(y, m, d, h, mn, 0) < datetime.now():
                expired.append(key)

        if expired:
            for key in expired:
                self.tasks.pop(key)
                self.tab.remove_task(key)
            self.tab.write_crontab()

        self.tab.clean_crontab(self.tasks.keys())

    def show_tasks(self):
        lgg.info('Show tasks: %s' % len(self.tasks))
        keys = sorted(self.tasks.keys())
        for key in keys:
            self.dialog.tasks_tree.set_single_task(key, self.tasks[key])

    def set_selection_mode(self):
        lgg.info('Enter in selection mode')
        self.core.is_interact_with_dialog = True
        self.dialog.hide()

    def call_settings_dialog(self):
        try:
            _ = self.video.category
            self.core.artelive.configure_downloading(self.video)
        except AttributeError:
            self.core.artetv.configure_downloading(self.video)

    def edit_tasks(self):
        self.tab = PyCronTab()
        isread = self.tab.read_crontab()
        if isread is not None:
            lgg.warning('Error when reading crontab:\n\t%s' % isread)

    def register_task(self, idx, task):
        lgg.info('Register task: %s' % idx)
        cmd = self.get_command(idx)
        self.tab.add_new_task(cmd, task['date'], task['time'])
        self.tab.write_crontab()
        self.dialog.tasks_tree.set_single_task(idx, task)

    def remove_task(self, idx):
        self.tasks.pop(idx)
        self.tab.remove_task(idx)
        self.tab.write_crontab()

    def get_command(self, index):
        """Create the command for crontab.

        """
        #'export DISPLAY=:0 & LC_CTYPE="<lang>" Lang="<lang>" 
        #                           qarte -a <ID> >> "<log>" 2>&1'
        #'export DISPLAY=:0 & LC_CTYPE="fr_BE.utf-8" Lang="fr_BE.utf-8" 
        #                           qarte -a 002 >> "path/to/log.log" 2>&1'
        lng, cod = locale.getdefaultlocale()
        lang = ".".join([lng, cod.lower()])
        return self.cron_cmd.replace('<lang>', lang)\
                            .replace('<ID>', index)\
                            .replace('<log>', self.logfile)


# Devel usage
class TaskEditor(object):
    def _write(self):
        def write(*args):
            self.tab.write()

        # To avoid an 'error: (4, 'System call interrupted')' the crontab
        # must be created in separate process
        Thread(target=write, args=(None,)).start()

class Core:
    def __init__(self):
        self.ui = None
        ws = os.path.expanduser("~") + "/qarte-3/wspace"
        self.workspace = {'user': ws}
        self.cfg = {'lang': 'fr',
                    'crontab_cmd': 'export DISPLAY=:0 & LC_CTYPE="<lang>" Lang="<lang>"'\
                                   ' <root> -a <ID> >> "<log>" 2>&1'}
        video = {'id': '047593-001-A', 'title': 'Mousson'}
        item = TVItem(video)
        ls = LoadingScheduler(self, item)

class TVItem:
    """Define a video item.

    """
    def __init__(self, attr, lang='fr'):
        self.idx = attr['id']
        self.title = attr['title']
        # Very short description
        #self.teaser = attr['teaser']
        #self.sched = attr['scheduled_on']
        #self.expire = self.format_expiration_date(attr['rights_end'])
        self.date = '2016-05-18 13:34:52' 
        self.duration = '43 min. 7 sec.'
        self.url = 'http://www.arte.tv/guide/fr/047593-001-A/mousson?autoplay=1'
        #self.subtitle = attr['subtitle']
        #self.adult = attr['adult']
        #self.pix_url = attr['thumbnail_url']
        #self.pixmap = None
        # description, origin and anno must be extracted from the video page
        #self.description = ''
        #self.origin = ''
        #self.anno = ''
        #self.views = attr['views']
        self.streams = {'HTTP_MP4_EQ_1': {'url': 'http://arte.gl-systemhaus.de/am/tvguide/ALL/066311-002-A_EQ_2_VOF-STF_02373307_MP4-1500_AMM-Tvguide.mp4'}}
        self.quality = 'HTTP_MP4_EQ_1'
        self.lang = lang
        self.set_file_name()

    def set_file_name(self):
        f = self.title.replace('/', '-').replace('»', '_').replace('«', '_')
        self.fname = f.replace('\\', ' ')


if __name__ == '__main__':
    app = QApplication(sys.argv)
    c = Core()


